# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: http://foglamp.readthedocs.io/
# FOGLAMP_END

""" Coral image classification poll plugin """

import copy
import json
import uuid
import logging
import time
import re
import os
from PIL import Image

from foglamp.common import logger
from foglamp.plugins.common import utils
from edgetpu.classification.engine import ClassificationEngine

__author__ = "Amandeep Singh Arora"
__copyright__ = "Copyright (c) 2019 Dianomic Systems Inc."
__license__ = "Apache 2.0"
__version__ = "${VERSION}"

_DEFAULT_CONFIG = {
    'plugin': {
        'description': 'Image classification using mobilenet v2 model',
        'type': 'string',
        'default': 'coral_classify',
        'readonly': 'true'
    },
    'model_file': {
        'description': 'TFlite model file to use for inference',
        'type': 'string',
        'default': 'mobilenet_v2_1.0_224_quant_edgetpu.tflite',
        'order': '1',
        'displayName': 'TFlite model file'
    },
    'labels_file': {
        'description': 'Labels file used during inference',
        'type': 'string',
        'default': 'imagenet_labels.txt',
        'order': '2',
        'displayName': 'Labels file'
    },
    'assetName': {
        'description': 'Asset Name',
        'type': 'string',
        'default': 'CoralClassify',
        'order': '3',
        'displayName': 'Asset Name'
    },
    'image_file': {
        'description': 'Image file to classify',
        'type': 'string',
        'default': '/tmp/cap.jpg',
        'order': '4',
        'displayName': 'Image file path'
    }
}
_LOGGER = logger.setup(__name__, level=logging.INFO)

engine = None
labels = None
asset_name = None
threshold = None
top_k = None


def load_labels(path):
    p = re.compile(r'\s*(\d+)(.+)')
    with open(path, 'r', encoding='utf-8') as f:
        lines = (p.match(line).groups() for line in f.readlines())
        return {int(num): text.strip() for num, text in lines}


def generate_svg(dwg, text_lines):
    for y, line in enumerate(text_lines):
        dwg.add(dwg.text(line, insert=(11, y*20+1), fill='black', font_size='20'))
        dwg.add(dwg.text(line, insert=(10, y*20), fill='white', font_size='20'))


def plugin_info():
    """ Returns information about the plugin.
    Args:
    Returns:
        dict: plugin information
    Raises:
    """

    return {
        'name': 'Imagenet classification plugin',
        'version': '1.7.0',
        'mode': 'poll',
        'type': 'south',
        'interface': '1.0',
        'config': _DEFAULT_CONFIG
    }


def plugin_init(config):
    """ Initialise the plugin that will identify/classify the object in image
        Args:
           config: JSON configuration document for the South plugin configuration category
        Returns:
           data: JSON object to be used in future calls to the plugin
        Raises:
    """
    global engine, labels, asset_name, threshold, top_k
    _LOGGER.debug("plugin_init: config = %s", json.dumps(config))
    try:
        model = config['model_file']['value']
        labels = config['labels_file']['value']
        asset_name = config['assetName']['value']
        threshold = 0.1
        top_k = 3

        model = os.path.join(os.path.dirname(__file__), "model", model)
        labels = os.path.join(os.path.dirname(__file__), "model", labels)

        _LOGGER.debug("Loading {} with {} labels.".format(model, labels))
        engine = ClassificationEngine(model)
        labels = load_labels(labels)

    except Exception as e:
        _LOGGER.exception("Imagenet classification plugin failed to start. Details %s", str(e))
        raise

    data = copy.deepcopy(config)
    return data


def plugin_poll(handle):
    """ Run inference on given image and returns a reading object in a JSON document.
    Available for poll mode only.
    Args:
        handle: handle returned by the plugin initialisation call
    Returns:
        returns a inference based reading in a JSON document, if it is available
        None - If no reading is available
    Raises:
        Exception
    """
    global engine, labels, asset_name, threshold, top_k
    try:
        img = Image.open(handle['image_file']['value'])

        start_time = time.monotonic()

        # ClassifyWithImage already resizes the image as per input tensor shape
        results = engine.ClassifyWithImage(img, threshold=threshold, top_k=top_k)
        end_time = time.monotonic()
        _LOGGER.debug('Inference time: {:0.2f} ms'.format((end_time - start_time) * 1000))
        '''
        text_lines = [
            'Inference time: %.2f ms' % ((end_time - start_time) * 1000),
        ]
        '''
        result_list = []
        for index, score in results:
            result_list.append({'label': labels[index], 'score': '%.2f' % score})

        data = {
            'asset': asset_name,
            'timestamp': utils.local_timestamp(),
            'key': str(uuid.uuid4()),
            'readings': {'result': result_list}
        }
        _LOGGER.debug(json.dumps(data))

    except (Exception, RuntimeError, FileNotFoundError) as ex:
        _LOGGER.exception("Coral image classification exception: {}".format(str(ex)))
        return None
    else:
        return data


def plugin_reconfigure(handle, new_config):
    """ Reconfigures the plugin

    it should be called when the configuration of the plugin is changed during the operation of the south service.
    The new configuration category should be passed.

    Args:
        handle: handle returned by the plugin initialisation call
        new_config: JSON object representing the new configuration category for the category
    Returns:
        new_handle: new handle to be used in the future calls
    Raises:
    """
    _LOGGER.info("Old config for Imagenet classification plugin {} \n new config {}".format(handle, new_config))

    plugin_shutdown(handle)
    new_handle = copy.deepcopy(new_config)
    plugin_init(new_handle)
    return new_handle


def plugin_shutdown(handle):
    global engine, labels, asset_name, threshold, top_k
    try:
        _LOGGER.info('Imagenet classification plugin shutting down.')
        engine = None
        labels = None
        asset_name = None
        threshold = None
        top_k = None
    except Exception as e:
        _LOGGER.exception(str(e))
        raise
